/*
 * (C) IBM International Business Machines (2002)
 * (A) David Oleszkiewicz 
 * (A) Daniel Stekloff
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*  We take a cpu on the command line and align ourselve to run on it.
 *  We do a loop of heavy math and sleeping, so that you should be able to 
 *  see this program on your running /usr/bin/top in another term.  I recommend
 *  top -d .5 for quick updates.
 */
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <math.h>
#include <getopt.h>
#include <fcntl.h>
#include "cpu_affinity.h"
#include "affinity.h"

#define PI 3.1419

extern int optind;                  /* for use with getopt */
extern char *optarg;                /* for use with getopt */
pid_t pid = 0;                      /* 0 is me */
unsigned long cpu_mask = 0x0;
struct sigaction handler; 
typedef enum {GET, SET, EXCLUSIVE} action_t;
int busy = 0;
int reset = 0;
action_t action = GET;
static struct option long_options[] = {
        {"any", 0, NULL, 'a'},      /* set to run on any proc */
        {"get", 0, NULL, 'g'},      /* get cpus_allowed */
        {"help", 0, NULL, 'h'},     /* print a usage help */
        {"busy", 0, NULL, 'b'},     /* busy loop */
        {"exclusive", 1, NULL, 'x'},/* set run_only */
        {"pid", 1, NULL, 'p'},      /* pid of process to control */
        {"reset", 0, NULL, 'r'},    /* reset all tasks' cpus_allowed */
        {"set", 1, NULL, 's'},      /* set cpus_allowed */
        {0, 0, 0, 0}
};
/* array of command line letter options */
const char options[] = "abhp:s:x:gr";

void usage(char *name)
{
        fprintf(stdout, "usage: %s <options>\n", name);
        fprintf(stdout, "\t-a set this process to run on any processor\n");
	fprintf(stdout, "\t-h print this help\n");
	fprintf(stdout, "\t-p <pid> specify the pid to work with (unspecified is self)\n");
	fprintf(stdout, "\t-x <cpu> set <pid> to run exclusively on <cpu>\n");
	fprintf(stdout, "\t-g get pid's cpu_mask (default behavior)\n");
	fprintf(stdout, "\t-s <cpu> set <pid> to run on <cpu>\n");
}

void parse_cmd_line(int argc, char *argv[])
{
        int c;
        while (1) {
                int option_index = 0;

                if ((c =
                     getopt_long(argc, argv, options, long_options,
                                 &option_index)) == -1) {
                        /* there are no more cmd line options to parse */
                        break;
                }

                switch (c) {
                case 'a': 
                        cpu_mask = ~(0UL); 
                        action = SET; 
                        break;
                case 'b': busy = 1; break;
                case 'g': action = GET; break;
                case 'h': usage(argv[0]); exit(0); break;
                case 'p': pid = atoi(optarg); break;
                case 'r':
                        fprintf(stderr, "reseting...\n");
                        sched_reset();
                        exit(0);
                        break;
                case 's': 
                        action = SET; 
                        cpu_mask = 1UL << atoi(optarg); 
                        break;
                case 'x': 
                        action = EXCLUSIVE; 
                        cpu_mask = 1UL << atoi(optarg); 
                        break;
                default: usage(argv[0]); exit(1); break;
                }
        }
}


void loop(void) 
{
	fprintf(stderr, "running in busy loop...");
        while (1) {
		int i;
		int res = 0;
                
		for (i = 0; i<400000; i++) {
                        res = res + (res * sin(i*(2*PI)))/i;
		}
		sleep(1);
        }
}

void signal_handler(int signal, siginfo_t * sip, void *notused) 
{
        switch (signal) {
        case SIGINT: {
                unsigned long mask;
                sched_get_affinity(getpid(), &mask);
                fprintf(stdout, "cpus_allowed = 0x%lx", mask);
                exit(1);
                break;
        }
        default:
        }
}
  
void install_signal_handler(void)
{
        /* sig handler takes 3 args | restart syscalls if possible */
        handler.sa_flags = SA_SIGINFO | SA_RESTART;
        handler.sa_sigaction = signal_handler;
        sigaction(SIGINT, &handler, NULL);
        return;
}



int main(int argc, char *argv[]) 
{
        int rc;

        pid = getpid();
        parse_cmd_line(argc, argv);
        install_signal_handler();

	fprintf(stdout, "pid = %d\n", getpid());

        switch (action) {
        case GET: {
                unsigned long mask;
                fprintf(stdout, "[get] pid:%d\n", pid);
                rc = sched_get_affinity(pid, &mask);
                fprintf(stdout, "current mask = 0x%lx\n", mask);
                break;
        }
        case SET: {
                unsigned long mask;
                fprintf(stdout, "[set] pid:%d mask:0x%lx\n", pid, cpu_mask);
                rc = sched_get_affinity(pid, &mask);
                fprintf(stdout, "current mask = 0x%lx\n", mask);
                rc = sched_set_affinity(pid, &cpu_mask);
                rc = sched_get_affinity(pid, &mask);
                fprintf(stdout, "new mask = 0x%lx\n", mask);
                break;
        }
        case EXCLUSIVE: {
                unsigned long mask;
                fprintf(stdout, "[exclusive] pid:%d mask:0x%lx\n", 
                        pid, cpu_mask);
                rc = sched_get_affinity(pid, &mask);
                fprintf(stdout, "current mask = 0x%lx\n", mask);
                rc = sched_set_exclusive(pid, &cpu_mask);
                rc = sched_get_affinity(pid, &mask);
                fprintf(stdout, "new mask = 0x%lx\n", mask);
                break;
        }
        default:
                usage(argv[0]);
                exit(1);
                break;
        }
        if (busy) loop();
        return 0;
}










