/*
 * (C) IBM International Business Machines (2002)
 * (A) David Oleszkiewicz 
 * (A) Daniel Stekloff
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <sys/syscall.h>
#include "config.h"
#include "cpu_affinity.h"
#include "kdebug.h"
#include "klog.h"

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
int init_module(void);
void cleanup_module(void);

int affinity_open(struct inode *inode, struct file *file);
int affinity_release(struct inode *inode, struct file *file);
int affinity_ioctl(struct inode *inode, struct file *filp,
                   unsigned int cmd, unsigned long arg);

static int affinity_major = 0;
MODULE_PARM(affinity_major, "i");   /* our chr_dev major number */

struct file_operations affinity_fops = {
        owner: THIS_MODULE,
        ioctl: affinity_ioctl,
};

/* ------------------------------------------------------------------------- */

static int affinity_ioctl(struct inode *inode, struct file *filp,
                   unsigned int cmd, unsigned long arg)
{
        int err = 0;
        int ret = 0;

        /* extract the type and number bitfields, and don't decode
         * wrong cmds: return ENOTTY before verify_area() */
        if (_IOC_TYPE(cmd) != AFFINITY_IOC_MAGIC) return -ENOTTY;
        if (_IOC_NR(cmd) > AFFINITY_IOC_MAXNR) return -ENOTTY;
        
        /* the direction is a bitmask, and VERIFY_WRITE catches R/W
         * transfers. `Type' is user-oriented, while
         * verify_area is kernel-oriented, so the concept of "read" and
         * "write" is reversed */
        if (_IOC_DIR(cmd) & _IOC_READ) {
                err = verify_area(VERIFY_WRITE, (void *)arg, _IOC_SIZE(cmd));
        } else if (_IOC_DIR(cmd) & _IOC_WRITE) {
                err =  verify_area(VERIFY_READ, (void *)arg, _IOC_SIZE(cmd));
        }
        if (err) return err;
        
        switch (cmd) {
        case AFFINITY_IOC_GET_AFFINITY: {
                struct affinity a;
                struct task_struct *task;

                a.pid = 45;
                a.mask = 0x1110;
                copy_from_user((void *)&a, (void *)arg, 
                               sizeof(struct affinity));
                kdebug("pid:%d mask:0x%lx", a.pid, a.mask);
                read_lock(&tasklist_lock);
                if (!(task = find_task_by_pid(a.pid))) {
                        read_unlock(&tasklist_lock);
                        return -ESRCH;
                }
                a.mask = task->cpus_allowed;
                read_unlock(&tasklist_lock);
                ret = copy_to_user((void *)arg, (void *)&a, 
                                   sizeof(struct affinity));
                break;
        }
        case AFFINITY_IOC_SET_AFFINITY: {
                struct affinity a;
                struct task_struct *task;
                
                copy_from_user((void *)&a, (void *)arg, sizeof(a));
                kdebug("a.pid = %d a.mask = 0x%lx", a.pid, a.mask);
                if (!(a.mask & cpu_online_map)) return -EINVAL;
                read_lock(&tasklist_lock);
                if (!(task = find_task_by_pid(a.pid))) {
                        read_unlock(&tasklist_lock);
                        return -ESRCH;
                }
                if ((current->euid != task->euid) 
                    && (current->euid != task->uid) 
                    && !capable(CAP_SYS_NICE)) {
                        read_unlock(&tasklist_lock);
                        return -EPERM;
                }
                task->cpus_allowed = a.mask;
                if (a.mask & (1UL << task->processor)) {
                        task->need_resched = 1;
                        read_unlock(&tasklist_lock);
                        schedule();
                } else {
                        read_unlock(&tasklist_lock);
                }
                ret = 0;
                break;
        }
        case AFFINITY_IOC_SET_EXCLUSIVE: {
                struct affinity a;
                struct task_struct *task;

                /* need to be root to do this */
                if (current->euid != 0) return -EPERM;
                copy_from_user((void *)&a, (void *)arg, sizeof(a));
                if (!(a.mask & cpu_online_map)) return -EINVAL;
                read_lock(&tasklist_lock);
                if (!(task = find_task_by_pid(a.pid))) {
                        read_unlock(&tasklist_lock);
                        return -ESRCH;
                }
                if ((current->euid != task->euid) 
                    && (current->euid != task->uid) 
                    && !capable(CAP_SYS_NICE)) {
                        read_unlock(&tasklist_lock);
                        return -EPERM;
                }
                task->cpus_allowed = a.mask;
                task->need_resched = 1;
                for_each_task(task) {
                        task->cpus_allowed = ~a.mask;
                }
                read_unlock(&tasklist_lock);
                ret = 0;
                break;
        }
        case AFFINITY_IOC_RESET: {
                struct task_struct *task;
                
                if (current->euid != 0) return -EACCES;
                read_lock(&tasklist_lock);
                for_each_task(task) {
                        task->cpus_allowed = ~(0UL);
                }
                read_unlock(&tasklist_lock);
                break;
        }
        case AFFINITY_IOC_GET_PROCESSOR: {
                struct affinity a;
                struct task_struct *task;

                copy_from_user((void *)&a, (void *)arg, 
                               sizeof(struct affinity));
                read_lock(&tasklist_lock);
                if (!(task = find_task_by_pid(a.pid))) {
                        read_unlock(&tasklist_lock);
                        return -ESRCH;
                }
                a.mask = task->processor;
                read_unlock(&tasklist_lock);
                ret = copy_to_user((void *)arg, (void *)&a, 
                                   sizeof(struct affinity));  
                break;
        }

        default: /* redundant, as cmd was checked against MAXNR */
                return -ENOTTY;
        }
        return ret;
}

static int __init affinity_init(void)
{
        int rc; 
        
        rc = register_chrdev(affinity_major, PACKAGE, &affinity_fops);
        if (rc < 0) {
		klog("error getting major number %d", affinity_major);
		return rc;
	}
	if (affinity_major == 0) {
                affinity_major = rc;
        }
        return 0;
}

static void __exit affinity_cleanup(void)
{
        int rc; 

	if ((rc = unregister_chrdev(affinity_major, PACKAGE)) < 0) {
                klog("error in unregister_chrdev: %d\n", rc);
        }
}

module_init(affinity_init);
module_exit(affinity_cleanup);

